# -*- coding: utf-8 -*-
import os

import yaml
import json
import posixpath

from six.moves.urllib import parse

from scalrctl import click, commands, defaults, settings, request
from scalrctl.commands.internal import bash_complete, update

__author__ = 'Dmitriy Korsakov, Sergey Babak'


CONFIGURATIONS = {
    'USER': {
        'API_HOST': {
            'order': 0,
            'description': 'Scalr API host'
        },
        'API_SCHEME': {
            'order': 1,
            'description': 'Scalr API scheme',
            'enum': ['http', 'https'],
        },
        'SSL_VERIFY_PEER': {
            'order': 2,
            'dependencies': {
                'API_SCHEME': 'https',
            },
            'description': 'SSL verification',
        },
        'API_KEY_ID': {
            'order': 3,
            'description': 'Scalr API key ID'
        },
        'API_SECRET_KEY': {
            'order': 4,
            'description': 'Scalr API secret key ID'
        },
        'envId': {
            'order': 5,
            'description': 'Scalr environment ID'
        },
        'accountId': {
            'order': 6,
            'description': 'Scalr account ID',
        },
        'view': {
            'order': 9,
            'description': 'View mode',
            'enum': ['tree', 'table', 'json']
        },
        'colored_output': {
            'order': 10,
            'description': 'Colored output'
        },
    },
    'ADMIN': {
        'GLOBAL_SCOPE_API_KEY_ID': {
            'order': 7,
            'description': 'Scalr admin (global scope) API key ID'
        },
        'GLOBAL_SCOPE_API_SECRET_KEY': {
            'order': 8,
            'description': 'Scalr admin (global scope) API secret key'
        },
    }
}


def _read_config(conf_path):
    if os.path.exists(conf_path):
        return yaml.load(open(conf_path, 'r'))


def _write_config(conf_path, conf_data):

    if not os.path.exists(defaults.CONFIG_DIRECTORY):
        os.makedirs(defaults.CONFIG_DIRECTORY)

    raw_data = yaml.dump(conf_data, default_flow_style=False, default_style='')
    with open(conf_path, 'w') as fp:
        fp.write(raw_data)


class ConfigureScalrCTL(commands.BaseAction):

    def run(self, *args, **kwargs):
        configure(**kwargs)

    def get_description(self):
        return "Set user configuration options in interactive mode"

    def get_options(self):
        profile_arg = click.Argument(('profile',), required=False)   # [ST-30]
        hlp = 'Add API credentials for Global scope'
        admin_opt = click.Option(('--with-global-scope', 'admin'), required=False, is_flag=True, help=hlp)
        return [profile_arg, admin_opt]


def configure(profile=None, admin=False):
    """
    Configure command-line client.
    Creates new profile in configuration directory
    and downloads spec files.
    :param profile: profile name
    :param admin: configure admin(global scope) values
    """

    conf_path = defaults.CONFIG_PATH if not profile else \
        os.path.join(defaults.CONFIG_DIRECTORY, '{}.yaml'.format(profile))

    conf_data = _read_config(conf_path) or {}

    values = CONFIGURATIONS['USER']
    if admin:
        values.update(CONFIGURATIONS['ADMIN'])

    click.echo('Configuring profile "{}":\n'.format(profile or 'default'))

    for key, value in sorted(values.items(), key=lambda kv: kv[1]['order']):

        default_value = getattr(settings, key)

        if key == "accountId" and not default_value:
            default_value = get_default_account_id(get_session_data(conf_data))

        desc = value.get('description') or key
        deps = value.get('dependencies')
        enum = value.get('enum')

        if deps and not all(conf_data[k] == v for k, v in deps.items()):
            continue

        if enum:
            desc = '{}. Choose from {}'.format(desc, ', '.join(enum))

        def _input():
            if type(default_value) in (bool,):
                conf_data[key] = click.confirm(desc, default=default_value)
            elif type(default_value) in (int, str) or default_value is None:
                conf_data[key] = click.prompt(desc, default=default_value)
                conf_data[key] = str(conf_data[key]).strip()

            return conf_data.get(key) in enum if enum else True

        while not _input():
            continue

        if key == "API_HOST":
            split_result = parse.urlsplit(conf_data[key])
            supported_schemas = values['API_SCHEME']['enum']
            if split_result.scheme and split_result.scheme in supported_schemas:
                url = split_result.netloc
                if split_result.path:
                    url_path = split_result.path[1:] if split_result.path.startswith('/') else split_result.path
                    url = posixpath.join(url, url_path)
                conf_data['API_HOST'] = url
                setattr(settings, 'API_SCHEME', split_result.scheme)

    _write_config(conf_path, conf_data)
    click.echo('\nNew config saved to {}\n'.format(conf_path))

    apply_settings(conf_data)
    update.update()
    bash_complete.setup_bash_complete()


def apply_settings(data):
    for key, value in data.items():
        if hasattr(settings, key):
            setattr(settings, key, value)


def get_session_data(data):
    try:
        api_level = (data['API_KEY_ID'], data['API_SECRET_KEY'])
        uri = '/api/%s/session/' % data['API_VERSION']
        raw_result = request.request(method="get", api_level=api_level, request_uri=uri)
        result = json.loads(raw_result)
    except:
        result = {}
    return result


def get_default_account_id(session_data):
    try:
        account_id = session_data["data"]['environments'][0]['accountId']
    except KeyError:
        account_id = None
    return account_id



__author__ = 'Dmitriy Korsakov'

import os

CONFIG_DIRECTORY = os.path.expanduser(os.environ.get("SCALRCLI_HOME", os.path.join(os.path.expanduser("~"), ".scalr")))
CONFIG_PATH = os.path.join(CONFIG_DIRECTORY, "%s.yaml" % os.environ.get("SCALRCLI_PROFILE", "default"))
VERSION = open(os.path.join(os.path.dirname(__file__), "VERSION")).read().strip()

API_LEVELS = [
    'user',
    'account',
    'global',
]



__author__ = 'Dmitriy Korsakov'
__doc__ = 'Manage global variables for servers'


from scalrctl import commands


class UpdateServerGlobalVariable(commands.Action):
    prompt_for = ["serverId", "globalVariableName"]



# -*- coding: utf-8 -*-
import json
import os
import pydoc
import sys

import yaml

from scalrctl import click, defaults, settings
from scalrctl.commands import Action
from scalrctl.commands.internal import configure, update

__author__ = 'Dmitriy Korsakov, Sergey Babak'


SCHEME_PATH = os.path.join(os.path.dirname(__file__), 'scheme/scheme.json')


if not os.path.exists(defaults.CONFIG_DIRECTORY):
    os.makedirs(defaults.CONFIG_DIRECTORY)

if os.path.exists(defaults.CONFIG_PATH):
    config_data = yaml.load(open(defaults.CONFIG_PATH, 'r'))
    configure.apply_settings(config_data)

if update.is_update_required():
    update.update()  # [ST-53]


def dummy_run():
    raise click.ClickException("Not implemented in current API version")


class ScalrCLI(click.Group):

    def __init__(self, name=None, commands=None, **attrs):
        if 'scheme' in attrs:
            self.scheme = attrs.pop('scheme')
        else:
            with open(SCHEME_PATH) as fp:
                self.scheme = json.load(fp)
        super(ScalrCLI, self).__init__(name, commands, chain=True, **attrs)

    def list_commands(self, ctx):
        """
        Returns a list of subcommand names.
        """

        return [k for k, v in self.scheme.items() if type(v) == dict]

    def get_cmd_groups(self):
        """
        List commands divided into groups.
        """

        groups = {}
        for cmd_name in self.scheme.keys():

            if cmd_name in ('cmd-group', 'group_descr', 'api_level'):
                continue

            if 'cmd-group' in self.scheme[cmd_name]:
                group_name = self.scheme[cmd_name]['cmd-group']
            elif 'api_level' in self.scheme[cmd_name]:
                scope_name = self.scheme[cmd_name]['api_level'].capitalize()
                group_name = '{} Scope operations'.format(scope_name)
            else:
                group_name = 'Other API commands'

            if group_name not in groups:
                groups[group_name] = [cmd_name, ]
            else:
                groups[group_name].append(cmd_name)

        return groups

    def format_commands(self, ctx, formatter):
        sections = {}
        for section_name, section_items in self.get_cmd_groups().items():
            rows = []
            for subcommand in section_items:
                cmd = self.get_command(ctx, subcommand)
                if cmd and not cmd.hidden:
                    rows.append((subcommand, cmd.short_help or ''))
            sections[section_name] = rows
        for name, rows in sections.items():
            rows.sort()
            with formatter.section(name):
                formatter.write_dl(rows)

    def get_command(self, ctx, name):
        """
        Given a context and a command name, this returns
        a `Command` object if it exists or returns `ScalrCLI`.
        """

        if name in self.scheme:
            subscheme = self.scheme[name]
        else:
            click.echo('No such command: {}.'.format(name))
            ctx.exit()

        # action level
        if 'route' in subscheme and 'http-method' in subscheme:
            hidden = subscheme.get('hidden', False)
            cls = pydoc.locate(
                subscheme['class']
            ) if 'class' in subscheme else Action
            action = cls(name=name,
                         route=subscheme.get('route'),
                         http_method=subscheme.get('http-method'),
                         api_level=subscheme.get('api_level'))

            try:
                action.validate()
            except AssertionError:
                dummy_help = 'Not implemented in current API version'
                return click.Command(name, params=[], callback=dummy_run,
                                     short_help=dummy_help, hidden=hidden)

            msg = subscheme.get('cmd_descr') or action.get_description()
            options = action.modify_options(action.get_options())
            cmd = click.Command(name, params=options, callback=action.run,
                                short_help=msg, help=msg, hidden=hidden)
            if 'epilog' in subscheme:
                cmd.epilog = subscheme['epilog']
            elif action.epilog:
                cmd.epilog = action.epilog
            return cmd
        else:
            return ScalrCLI(scheme=subscheme,
                            short_help=subscheme.get('group_descr', ''))


@click.command(cls=ScalrCLI)
@click.version_option()
@click.pass_context
@click.option('--key_id', help="API key ID")
@click.option('--secret_key', help="API secret key")
@click.option('--config', help="Path to a custom scalr-ctl configuration file")
def cli(ctx, key_id, secret_key, config, *args, **kvargs):
    """Scalr-ctl is a command-line interface to your Scalr account"""

    service_cmd = any(arg in ('configure', 'update') for arg in sys.argv)

    if key_id:
        settings.API_KEY_ID = str(key_id)

    if secret_key:
        settings.API_SECRET_KEY = str(secret_key)
    elif (settings.API_KEY_ID and settings.API_KEY_ID.strip()
          and not settings.API_SECRET_KEY
          and not service_cmd):  # [ST-21]
        settings.API_SECRET_KEY = str(click.prompt(text='API SECRET KEY',
                                                   hide_input=True))
    if config:
        if os.path.exists(config):
            config_data = yaml.load(open(config, 'r'))
            configure.apply_settings(config_data)
        else:
            msg = 'Configuration file not found: {}'.format(config)
            raise click.ClickException(msg)


if __name__ == '__main__':
    cli()



__author__ = 'Dmitriy Korsakov'
__doc__ = 'Role management'

import copy
from scalrctl import commands
from scalrctl import click


class ChangeRoleAttrs(commands.Action):
    prompt_for = ["roleId"]


class RoleClone(commands.SimplifiedAction):

    epilog = "Example: scalr-ctl role clone --roleId <ID> --name MyNewRole"
    post_template = {
        "cloneRoleRequest": {"name": ""}
    }

    def get_options(self):
        hlp = "The name of a new Role."
        new_name = click.Option(('--name', 'name'), required=True, help=hlp)
        options = [new_name, ]
        options.extend(super(RoleClone, self).get_options())
        return options


    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        name = kwargs.pop("name", None)
        post_data = copy.deepcopy(self.post_template)
        post_data["cloneRoleRequest"]["name"] = name
        kv = {"import-data": post_data}
        kv.update(kwargs)
        arguments, kw = super(RoleClone, self).pre(*args, **kv)
        return arguments, kw


class RolePromote(commands.SimplifiedAction):

    epilog = "Example: scalr-ctl role promote --roleId <ID>"
    post_template = {
        "cloneRoleRequest": {"name": ""}
    }

    def get_options(self):
        hlp = "The name of a new Role."
        new_name = click.Option(('--name', 'name'), required=True, help=hlp)
        options = [new_name, ]
        options.extend(super(RoleClone, self).get_options())
        return options


    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        name = kwargs.pop("name", None)
        post_data = copy.deepcopy(self.post_template)
        post_data["cloneRoleRequest"]["name"] = name
        kv = {"import-data": post_data}
        kv.update(kwargs)
        arguments, kw = super(RoleClone, self).pre(*args, **kv)
        return arguments, kw


class RolePromote(commands.SimplifiedAction):

    epilog = "Example: scalr-ctl role promote --roleId <ID>"
    post_template = {}

    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        kv = {"import-data": {}}
        kv.update(kwargs)
        arguments, kw = super(RolePromote, self).pre(*args, **kv)
        return arguments, kw


__author__ = 'Dmitriy Korsakov'
__doc__ = 'Farm management'

import json
import copy

from scalrctl import commands
from scalrctl import click

from scalrctl import request, settings


class FarmTerminate(commands.SimplifiedAction):

    epilog = "Example: scalr-ctl farms terminate --farmId <ID> --force"

    post_template = {
        "terminateFarmRequest": {"force": True}
    }

    def get_options(self):
        hlp = "It is used to terminate the Server immediately ignoring scalr.system.server_terminate_timeout."
        force_terminate = click.Option(('--force', 'force'), is_flag=True, default=False, help=hlp)
        options = [force_terminate, ]
        options.extend(super(FarmTerminate, self).get_options())
        return options


    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        force = kwargs.pop("force", None)
        post_data = copy.deepcopy(self.post_template)
        post_data["terminateFarmRequest"]["force"] = force
        kv = {"import-data": post_data}
        kv.update(kwargs)
        arguments, kw = super(FarmTerminate, self).pre(*args, **kv)
        return arguments, kw


class FarmLaunch(commands.SimplifiedAction):

    epilog = "Example: scalr-ctl farms launch --farmId <ID>"
    post_template = {}

    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        kv = {"import-data": {}}
        kv.update(kwargs)
        arguments, kw = super(FarmLaunch, self).pre(*args, **kv)
        return arguments, kw


class FarmClone(commands.SimplifiedAction):

    epilog = "Example: scalr-ctl farms clone --farmId <ID> --name MyNewFarm"
    post_template = {
        "cloneFarmRequest": {"name": ""}
    }

    def get_options(self):
        hlp = "The name of a new Farm."
        hard_terminate = click.Option(('--name', 'name'), required=True, help=hlp)
        options = [hard_terminate, ]
        options.extend(super(FarmClone, self).get_options())
        return options


    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        name = kwargs.pop("name", None)
        post_data = copy.deepcopy(self.post_template)
        post_data["cloneFarmRequest"]["name"] = name
        kv = {"import-data": post_data}
        kv.update(kwargs)
        arguments, kw = super(FarmClone, self).pre(*args, **kv)
        return arguments, kw


class FarmSuspend(FarmLaunch):

    epilog = "Example: scalr-ctl farms suspend --farmId <ID>"
    post_template = {}


class FarmResume(FarmLaunch):

    epilog = "Example: scalr-ctl farms resume --farmId <ID>"
    post_template = {}


class FarmLock(commands.SimplifiedAction):

    epilog = "Example: scalr-ctl farm lock --farmId <ID> --comment <COMMENT> --unlock-permission <ANYONE|OWNER|TEAM>"

    post_template = {
        "lockFarmRequest": {"lockComment": "", "unlockPermission": "anyone"}
    }

    def get_options(self):
        comment = click.Option(('--lockComment', 'comment'), default="", help="Comment to lock a Farm.")
        hlp = "If you would like to prevent other users unlocking the Farm you should set 'owner' options.\
                  With 'team' options only members of the Farm's Teams can unlock this Farm.\
                  Default value 'anyone' means that anyone with access can unlock this Farm."
        unlock_permission = click.Option((
            '--unlockPermission', 'unlock_permission'),
            default="anyone", show_default=True, help=hlp)
        options = [comment, unlock_permission]
        options.extend(super(FarmLock, self).get_options())
        return options

    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        comment = kwargs.pop("comment", None)
        unlock_permission = kwargs.pop("unlock_permission", "anyone")
        post_data = copy.deepcopy(self.post_template)
        post_data["lockFarmRequest"]["lockComment"] = comment
        post_data["lockFarmRequest"]["unlockPermission"] = unlock_permission
        kv = {"import-data": post_data}
        kv.update(kwargs)
        arguments, kw = super(FarmLock, self).pre(*args, **kv)
        return arguments, kw


class FarmCreateFromTemplate(commands.Action):

    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        kwargs = self._apply_arguments(**kwargs)
        stdin = kwargs.pop('stdin', None)
        kwargs["FarmTemplate"] = self._read_object() if stdin else self._edit_example()
        return args, kwargs

    def run(self, *args, **kwargs):
        """
        Callback for click subcommand.
        """
        hide_output = kwargs.pop('hide_output', False)  # [ST-88]
        args, kwargs = self.pre(*args, **kwargs)

        uri = self._request_template
        payload = {}
        data = {}

        if '{envId}' in uri and not kwargs.get('envId') and settings.envId:
            kwargs['envId'] = settings.envId

        if kwargs:
            # filtering in-body and empty params
            uri = self._request_template.format(**kwargs)

            for key, value in kwargs.items():
                param = '{{{}}}'.format(key)
                if value and (param not in self._request_template):
                    data.update(value)

        if self.dry_run:
            click.echo('{} {} {} {}'.format(self.http_method, uri,
                                            payload, data))
            # returns dummy response
            return json.dumps({'data': {}, 'meta': {}})

        data = json.dumps(data)
        raw_response = request.request(self.http_method, self.api_level,
                                       uri, payload, data)
        response = self.post(raw_response)

        text = self._format_response(response, hidden=hide_output, **kwargs)
        if text is not None:
            click.echo(text)

        return response

    def _edit_example(self):
        commentary = \
            '''# The body must be a valid FarmTemplate object.
#
# Type your FarmTemplate object below this line. The above text will not be sent to the API server.'''
        text = click.edit(commentary)
        if text:
            raw_object = "".join([line for line in text.splitlines()
                                  if not line.startswith("#")]).strip()
        else:
            raw_object = ""
        return json.loads(raw_object)



__author__ = 'Dmitriy Korsakov'
__doc__ = 'Manage global variables for roles'


from scalrctl import commands


class UpdateRoleGlobalVariable(commands.Action):
    prompt_for = ["roleId", "globalVariableName"]

class DeleteRoleGlobalVariable(commands.Action):
    delete_target = 'globalVariableName'


import os

from distutils.core import setup
from distutils.command.install import install


try:
    from post_setup import main as post_install
except ImportError:
    post_install = lambda: None


class _install(install):
    def run(self):
        install.run(self)
        post_install()  # Does not work with pip


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

description = "Scalr-ctl is a command-line interface to your Scalr account"


if __name__ == '__main__':

    setup(
        name='scalr-ctl',
        version=read("scalrctl/VERSION").strip(),
        description=description,
        author="Scalr Inc.",
        author_email="dmitry@scalr.com",
        url="https://scalr.net",
        license="Apache-2",
        platforms="any",
        long_description=read('README'),
        classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: System Administrators',
            'Topic :: Utilities'
            ],
        packages=[
            "scalrctl",
            "scalrctl.click",
            "scalrctl.commands",
            "scalrctl.commands.internal"
        ],
        include_package_data=True,
        package_data={
            '': ['VERSION', 'scheme/scheme.json'],
        },
        data_files=[('', ['scalrctl/scheme/scheme.json', ]), ],
        install_requires=[
            'prettytable>=0.7.2',
            'pyyaml>=3.11',
            'requests>=2.10.0',
            'six>=1.10.0',
            'colorama>=0.3.7',
            'dicttoxml>=1.7.4',
        ],
        entry_points='''
            [console_scripts]
            scalr-ctl=scalrctl.app:cli
        ''',
    )



# -*- coding: utf-8 -*-
import json
import re
import os

import dicttoxml

from scalrctl import click, request, settings, utils, view, examples, defaults

__author__ = 'Dmitriy Korsakov'


class MultipleClickException(click.ClickException):

    def format_message(self):
        return '\x1b[31m%s\x1b[39m' % self.message if settings.colored_output else self.message

    def show(self, file=None):
        if file is None:
            file = click._compat.get_text_stderr()
        click.utils.echo('%s' % self.format_message(), file=file)


class BaseAction(object):

    epilog = None

    def __init__(self, *args, **kwargs):
        pass

    def run(self, *args, **kwargs):
        pass

    def get_description(self):
        return ''

    def modify_options(self, options):
        return options

    def get_options(self):
        return []

    def validate(self):
        pass


class Action(BaseAction):

    raw_spec = None

    # Optional. Some values like GCE imageId
    # cannot be passed through command lines
    prompt_for = None

    # Temporary. Object definitions in YAML
    # spec are not always correct
    mutable_body_parts = None

    # Optional, e.g. '#/definitions/GlobalVariable'
    object_reference = None

    dry_run = False
    post_template = None
    _table_columns = None

    _discriminators = {}

    ignored_options = ()
    delete_target = None

    def __init__(self, name, route, http_method, api_level, *args, **kwargs):
        self.name = name
        self.route = route
        self.http_method = http_method
        self.api_level = api_level
        self._table_columns = []

        self._init()

    def _init(self):
        self.raw_spec = utils.read_spec(self.api_level, ext='json')

        if not self.epilog and self.http_method.upper() == 'POST':
            msg = "Example: scalr-ctl {level} {name} < {name}.json"
            self.epilog = msg.format(
                level=self.api_level if self.api_level != 'user' else '',
                name=self.name
            )

    def _check_arguments(self, **kwargs):
        route_data = self.raw_spec['paths'][self.route]
        if 'parameters' not in route_data:
            return

        for param in route_data['parameters']:
            pattern = param.get('pattern')
            param_name = param.get('name')
            if pattern and param_name and param_name in kwargs:
                value = str(kwargs[param_name])
                matches = re.match(pattern, value.strip())
                if not matches or len(matches.group()) != len(value):
                    raise click.ClickException("Invalid value for {}"
                                               .format(param_name))

    def _apply_arguments(self, **kwargs):
        if kwargs.get('filters'):
            for pair in kwargs.pop('filters').split(','):
                kv = pair.split('=')
                if len(kv) == 2:
                    kwargs[kv[0]] = kv[1]

        if kwargs.get('columns'):
            self._table_columns = kwargs.pop('columns').split(',')

        if kwargs.pop('dryrun', False):
            self.dry_run = True

        if kwargs.pop('debug', None):
            settings.debug_mode = True

        if kwargs.pop('nocolor', None):
            settings.colored_output = False

        if kwargs.get('transformation'):
            settings.view = kwargs.pop('transformation')

        return kwargs

    def _get_object(self, *args, **kwargs):
        try:
            obj = self.__class__(name='get', route=self.route,
                                 http_method='get', api_level=self.api_level)
            raw_text = obj.run(*args, **kwargs)
            utils.debug(raw_text)
            if raw_text is None:
                return {}
            json_text = json.loads(raw_text)
            filtered = self._filter_json_object(json_text['data'],
                                                filter_createonly=True)
            return json.dumps(filtered, indent=2)
        except Exception as e:
            utils.reraise(e)

    def _edit_object(self, *args, **kwargs):
        raw_object = self._get_object(*args, **kwargs)
        raw_object = click.edit(raw_object)
        if raw_object is None:
            raise ValueError("No changes in JSON")
        return json.loads(raw_object)

    def _edit_example(self):
        commentary = examples.create_post_example(self.api_level, self.route)
        text = click.edit(commentary)
        if text:
            raw_object = "".join([line for line in text.splitlines()
                                  if not line.startswith("#")]).strip()
        else:
            raw_object = ""
        return json.loads(raw_object)

    @staticmethod
    def _read_object():
        """
        Reads JSON object from stdin.
        """
        raw_object = click.get_text_stream('stdin').read()
        return json.loads(raw_object)

    def _format_errmsg(self, errors):
        messages = []
        num = 1
        for error_data in errors:
            err_code = '%s:' % error_data['code'] if 'code' in error_data else ''
            err_msg = error_data.get('message', '')
            err_index = ':' if len(errors) == 1 else ' %s:' % num
            err_line = 'Error%s %s %s' % (err_index, err_code, err_msg)
            messages.append(err_line)
            num += 1
        result_errmsg = '\n'.join(messages)

        return result_errmsg

    def _format_response(self, response, hidden=False, **kwargs):
        text = None

        if response:
            try:
                response_json = json.loads(response)
            except ValueError:
                utils.debug("Server response: {}".format(str(response)))
                utils.reraise("Invalid server response")

            errors = response_json.get('errors')

            if errors:
                errmsg = self._format_errmsg(errors)
                error = MultipleClickException(errmsg)
                error.code = 1
                raise error

            if not hidden:
                utils.debug(response_json.get('meta'))

            if hidden:
                pass
            elif settings.view in ('raw', 'json'):
                click.echo(response)
            elif settings.view == 'xml':
                click.echo(dicttoxml.dicttoxml(response_json))
            elif settings.view == 'tree':
                data = json.dumps(response_json.get('data'))
                click.echo(view.build_tree(data))
            elif settings.view == 'table':
                columns = self._table_columns or self._get_column_names()
                rows, current_page, last_page = view.calc_table(response_json,
                                                                columns)
                pre = "Page: {} of {}".format(current_page, last_page)
                click.echo(view.build_table(columns, rows, pre=pre))  # XXX
        elif self.http_method.upper() == 'DELETE':
            deleted_id = kwargs.get(self.delete_target, '') or ''
            if not deleted_id:
                for param, value in kwargs.items():
                    if 'Id' in param and param not in ('envId', 'accountId'):
                        deleted_id = value
                        break
            if not deleted_id:
                for param, value in kwargs.items():
                    if 'Name' in param:
                        deleted_id = value
                        break
            text = "Deleted {}".format(deleted_id)
        return text

    def _get_default_options(self):
        options = []
        for param in self._get_raw_params():
            option = click.Option(('--{}'.format(param['name']),
                                  param['name']), required=param['required'],
                                  help=param['description'],
                                  default=param.get('default'),
                                  show_default='default' in param)
            options.append(option)
        return options

    def _get_custom_options(self):
        options = []

        if self.http_method.upper() in ('POST', 'PATCH'):
            stdin = click.Option(('--stdin', 'stdin'),
                                 is_flag=True, required=False,
                                 help="Read JSON data from standart console "
                                      "input instead of editing it in the "
                                      "default console text editor.")
            options.append(stdin)
            """
            interactive = click.Option(('--interactive', 'interactive'),
                                       is_flag=True, required=False,
                                       help="Edit JSON data in the default "
                                            "console text editor before "
                                            "sending POST request to server.")
            options.append(interactive)
            """

        if self.http_method.upper() == 'GET':
            if self._returns_iterable():
                maxres = click.Option(('--max-results', 'maxResults'),
                                      type=int, required=False,
                                      help="Maximum number of records. "
                                      "Example: --max-results=2")
                options.append(maxres)

                pagenum = click.Option(('--page-number', 'pageNum'), type=int,
                                       required=False, help="Current page "
                                       "number. Example: --page-number=3")
                options.append(pagenum)

                filters = self._get_available_filters()
                if filters:
                    filters = sorted(filters)
                    filter_help = ("Apply filters. Example: type=ebs,size=8."
                                   "Available filters: {}."
                                   ).format(', '.join(filters))
                    filters = click.Option(('--filters', 'filters'),
                                           required=False, help=filter_help)
                    options.append(filters)

                columns_help = ("Filter columns in table view "
                                "[--table required]. Example: NAME,SIZE,"
                                "SCOPE. Available columns: {}."
                                ).format(', '.join(self._get_column_names()))
                columns = click.Option(('--columns', 'columns'),
                                       required=False, help=columns_help)
                options.append(columns)

            raw = click.Option(('--raw', 'transformation'), is_flag=True,
                               flag_value='raw', default=False, hidden=True,
                               help="Print raw response")
            json_ = click.Option(('--json', 'transformation'), is_flag=True,
                                 flag_value='raw', default=False,
                                 help="Print raw response")
            xml = click.Option(('--xml', 'transformation'), is_flag=True,
                               flag_value='xml', default=False,
                               help="Print response as a XML")
            tree = click.Option(('--tree', 'transformation'), is_flag=True,
                                flag_value='tree', default=False,
                                help="Print response as a colored tree")
            nocolor = click.Option(('--nocolor', 'nocolor'), is_flag=True,
                                   default=False, help="Use colors")
            options += [raw, tree, nocolor, json_, xml]

            if self.name not in ('get', 'retrieve'):  # [ST-54] [ST-102]
                table = click.Option(('--table', 'transformation'),
                                     is_flag=True, flag_value='table',
                                     default=False,
                                     help="Print response as a colored table")
                options.append(table)

        debug = click.Option(('--debug', 'debug'), is_flag=True,
                             default=False, help="Print debug messages")
        options.append(debug)

        return options

    def _get_body_type_params(self):
        route_data = self.raw_spec['paths'][self.route][self.http_method]
        return [param for param in route_data.get('parameters', '')]

    def _get_path_type_params(self):
        route_data = self.raw_spec['paths'][self.route]
        return [param for param in route_data.get('parameters', '')]

    def _get_raw_params(self):
        result = self._get_path_type_params()
        if self.http_method.upper() in ('GET', 'DELETE'):
            body_params = self._get_body_type_params()
            result.extend(body_params)
        return result

    def _returns_iterable(self):
        responses = self.raw_spec['paths'][self.route]['get']['responses']
        if '200' in responses:
            response_200 = responses['200']
            if 'schema' in response_200:
                schema = response_200['schema']
                if '$ref' in schema:
                    object_key = schema['$ref'].split('/')[-1]
                    object_descr = self.raw_spec['definitions'][object_key]
                    object_properties = object_descr['properties']
                    data_structure = object_properties['data']
                    return 'array' == data_structure.get('type')
        return False

    def _get_available_filters(self):
        if self._returns_iterable():
            data = self._result_descr['properties']['data']
            response_ref = data['items']['$ref']
            response_descr = self._lookup(response_ref)
            if 'x-filterable' in response_descr:
                return response_descr['x-filterable']
        return []

    def _get_column_names(self):
        data = self._result_descr['properties']['data']
        # XXX: Inconsistency in swagger spec.
        # See RoleDetailsResponse vs RoleCategoryListResponse
        response_ref = data['items']['$ref'] \
            if 'items' in data else data['$ref']
        response_descr = self._lookup(response_ref)
        properties = response_descr['properties']

        column_names = []
        for k, v in properties.items():
            if '$ref' not in v:
                column_names.append(k)
            else:  # ST-226
                f_key = self._lookup(v['$ref'])
                if "properties" in f_key:
                    if len(f_key["properties"]) == 1:
                        if "id" in f_key["properties"]:
                            if "type" in f_key["properties"]["id"]:
                                if f_key["properties"]["id"]["type"] in ("integer", "string"):
                                    column_names.append("%s.id" % k)

        return column_names

    def _lookup(self, response_ref):
        """
        Returns document section
        Example: #/definitions/Image returns Image defenition section.
        """
        if response_ref.startswith('#'):
            paths = response_ref.split('/')[1:]
            result = self.raw_spec
            for path in paths:
                if path not in result:
                    return
                result = result[path]
            return result

    @property
    def _result_descr(self):
        route_data = self.raw_spec['paths'][self.route]
        responses = route_data[self.http_method]['responses']
        if '200' in responses:
            response_200 = responses['200']
            if 'schema' in response_200:
                schema = response_200['schema']
                if '$ref' in schema:
                    response_ref = schema['$ref']
                    return self._lookup(response_ref)

    def _list_concrete_types(self, schema):
        types = []
        if "x-concreteTypes" in schema:
            for ref_dict in schema["x-concreteTypes"]:
                ref_link = ref_dict['$ref']
                types += [link.split("/")[-1] for link in self._list_concrete_types_recursive(ref_link)]
        return types

    def _list_concrete_types_recursive(self, reference):
        references = []
        schema = self._lookup(reference)
        if "x-concreteTypes" not in schema:
            references.append(reference)
        else:
            for ref_dict in schema["x-concreteTypes"]:
                references += self._list_concrete_types_recursive(ref_dict['$ref'])
        return references

    def _filter_json_object(self, data, filter_createonly=False,
                            schema=None, reference=None):
        """
        Removes immutable parts from JSON object
        before sending it in POST or PATCH.
        """
        filtered = {}

        # load `schema`
        if schema is None:
            for param in self._get_body_type_params():
                if 'schema' in param:
                    schema = param['schema']
                    if '$ref' in schema:
                        reference = schema['$ref']
                        schema = self._lookup(schema['$ref'])
                    break

        # load child object as `schema`
        if 'discriminator' in schema:

            disc_key = schema['discriminator']
            disc_path = '{}/{}'.format(reference, disc_key)
            disc_value = data.get(disc_key) or self._discriminators.get(disc_path)

            if not disc_value:
                raise click.ClickException((
                    "Provided JSON object is incorrect: missing required param '{}'."
                ).format(disc_key))
            elif disc_value not in self._list_concrete_types(schema):
                raise click.ClickException((
                    "Provided JSON object is incorrect: required "
                    "param '{}' has invalid value '{}', must be one of: {}."
                ).format(disc_key, disc_value, self._list_concrete_types(schema)))
            else:
                # save discriminator for current reference/key
                self._discriminators[disc_path] = disc_value

            reference = '#/definitions/{}'.format(disc_value)
            schema = self._lookup(reference)

        # filter input data by properties of `schema`
        if schema and 'properties' in schema:
            create_only_props = schema.get('x-createOnly', '')
            for p_key, p_value in schema['properties'].items():

                if reference:
                    key_path = '.'.join([reference.split('/')[-1], p_key])
                else:
                    key_path = p_key

                if p_key not in data:
                    utils.debug("Ignore {}, unknown key.".format(key_path))
                    continue
                if p_value.get('readOnly'):
                    utils.debug("Ignore {}, read-only key.".format(key_path))
                    continue
                if filter_createonly and p_key in create_only_props:
                    utils.debug("Ignore {}, create-only key.".format(key_path))
                    continue

                if '$ref' in p_value and isinstance(data[p_key], dict):
                    # recursive filter sub-object
                    utils.debug("Filter sub-object: {}.".format(p_value['$ref']))
                    filtered[p_key] = self._filter_json_object(
                        data[p_key],
                        filter_createonly=filter_createonly,
                        reference=p_value['$ref'],
                        schema=self._lookup(p_value['$ref']),
                    )
                else:
                    # add valid key-value
                    filtered[p_key] = data[p_key]

        return filtered

    def _list_createonly_properties(self):
        """
        Some properties (mostly IDs) cannot be marked as read-only because
        they are passed in PUT-methods (but still must be filtered in
        UPDATE-methods).
        """
        result_descr = self._result_descr
        if result_descr and 'properties' in result_descr:
            data = result_descr['properties']['data']
            if 'items' in data:
                path = data['items']['$ref']
            else:
                path = data['$ref']
            obj = self._lookup(path)
            return obj['x-createOnly'] if 'x-createOnly' in obj else []

    @property
    def _request_template(self):
        basepath_uri = self.raw_spec['basePath']
        return '{}{}'.format(basepath_uri, self.route)

    def pre(self, *args, **kwargs):
        """
        Before request is made.
        """
        kwargs = self._apply_arguments(**kwargs)
        self._check_arguments(**kwargs)

        import_data = kwargs.pop('import-data', {})
        #interactive = kwargs.pop('interactive', None)
        stdin = kwargs.pop('stdin', None)
        http_method = self.http_method.upper()

        if http_method not in ('PATCH', 'POST'):
            return args, kwargs

        # prompting for body and then validating it
        for param_name in (p['name'] for p in self._get_body_type_params()):
            try:
                if param_name in import_data:
                    json_object = self._filter_json_object(
                        import_data[param_name],
                        filter_createonly=True
                    ) if http_method == 'PATCH' else import_data[param_name]
                else:
                    if http_method in ('POST', 'PATCH') and stdin:
                        json_object = self._read_object()
                    elif http_method == 'PATCH':
                        json_object = self._edit_object(*args, **kwargs)
                    elif http_method == 'POST':
                        json_object = self._edit_example()

                json_object = self._filter_json_object(json_object)
                kwargs[param_name] = json_object
            except ValueError as e:
                utils.reraise(e)

        return args, kwargs

    def post(self, response):
        """
        After request is made.
        """
        return response

    #
    # PUBLIC METHODS
    #

    def run(self, *args, **kwargs):
        """
        Callback for click subcommand.
        """
        hide_output = kwargs.pop('hide_output', False)  # [ST-88]
        args, kwargs = self.pre(*args, **kwargs)

        uri = self._request_template
        payload = {}
        data = {}

        if '{envId}' in uri and not kwargs.get('envId') and settings.envId:
            kwargs['envId'] = settings.envId

        if '{accountId}' in uri and not kwargs.get('accountId') and settings.accountId:
            kwargs['accountId'] = settings.accountId

        if kwargs:
            # filtering in-body and empty params
            uri = self._request_template.format(**kwargs)
            for key, value in kwargs.items():
                param = '{{{}}}'.format(key)
                if value and (param not in self._request_template):
                    body_params = self._get_body_type_params()
                    if self.http_method.upper() in ('GET', 'DELETE'):
                        payload[key] = value
                    elif body_params and key == body_params[0]['name']:
                        data.update(value)

        if self.dry_run:
            click.echo('{} {} {} {}'.format(self.http_method, uri,
                                            payload, data))
            # returns dummy response
            return json.dumps({'data': {}, 'meta': {}})

        data = json.dumps(data)
        raw_response = request.request(self.http_method, self.api_level,
                                       uri, payload, data)
        response = self.post(raw_response)

        text = self._format_response(response, hidden=hide_output, **kwargs)
        if text is not None:
            click.echo(text)

        return response

    def get_description(self):
        """
        Returns action description.
        """
        route_data = self.raw_spec['paths'][self.route]
        return route_data[self.http_method]['description']

    def modify_options(self, options):
        """
        This is the place where command line options can be fixed
        after they are loaded from yaml spec.
        """
        for option in options:
            if self.prompt_for and option.name in self.prompt_for:
                option.prompt = option.name
            if (option.name == 'envId' and settings.envId) or \
                    (option.name == 'accountId' and settings.accountId):
                option.required = False
        return options

    def get_options(self):
        """
        Returns action options.
        """
        options = self._get_default_options() + self._get_custom_options()
        return [opt for opt in options if opt.name not in self.ignored_options]

    def validate(self):
        """
        Validate routes for current API scope.
        """
        if self.route and self.api_level:
            spec_path = os.path.join(defaults.CONFIG_DIRECTORY,
                                     '{}.json'.format(self.api_level))
            api_routes = json.load(open(spec_path, 'r'))['paths'].keys()
            try:
                assert api_routes and self.route in api_routes, self.name
            except AssertionError:  # ST-224
                if self.api_level == "account":
                    bc_route = self.route.replace('{accountId}/', '')
                    assert api_routes and bc_route in api_routes, self.name
                    self.route = bc_route
                else:
                    raise


class SimplifiedAction(Action):
    ignored_options = ('stdin',)
